package fr.lri.hcc.ilda.touchtokens;

/*
 *   Authors: Caroline Appert (caroline.appert@lri.fr)
 *   Copyright (c) Univ. Paris-Sud XI, 2015. All Rights Reserved
 *   Licensed under the GNU LGPL. For full terms see the file COPYING.
*/

import android.content.Context;
import android.graphics.Canvas;
import android.graphics.Color;
import android.graphics.Paint;
import android.support.v4.view.MotionEventCompat;
import android.util.AttributeSet;
import android.util.Log;
import android.view.MotionEvent;
import android.view.View;

import java.util.ArrayList;
import java.util.Timer;
import java.util.TimerTask;

/**
 * Created by appert on 02/09/15.
 */
public abstract class TouchTokenView extends View {

    protected String tokenDown = null;

    protected static TouchPoint[] FINGERS_LOCATION = new TouchPoint[10];
    protected static TouchPoint[] FINGERS_LOCATION_AT_TIMER_START_TIME = new TouchPoint[10];

    protected int dwellDuration = 200; // in ms
    protected double dwellTolerance = 3; // in mm

    protected Timer dwellTimer;
    protected long dwellStartTime = -1;
    protected long lastUpdateTime = -1;

    protected Paint tracePaint = new Paint();

    protected boolean debug = true;

    protected double viewWidthMm = -1;
    protected double viewHeightMm = -1;

    public TouchTokenView(Context context) {
        super(context);
        init();
    }

    public TouchTokenView(Context context, AttributeSet attrs) {
        super(context, attrs);
        init();
    }

    public TouchTokenView(Context context, AttributeSet attrs, int defStyleAttr) {
        super(context, attrs, defStyleAttr);
        init();
    }

    protected void init() {

        viewWidthMm = getWidth() / MainActivity.SCREEN_RESOLUTION;
        viewHeightMm = getHeight() / MainActivity.SCREEN_RESOLUTION;

        Log.v("TOUCHTOKENS", "viewWidthMm="+viewWidthMm+", viewHeightMm="+viewHeightMm);

        dwellTimer = new Timer();
        dwellTimer.schedule(new TimerTask() {
            public void run() {
                if(dwellStartTime == -1) {
                    return;
                }
                if(tokenDown != null) {
                    return;
                }
                long currentTime = System.currentTimeMillis();
                if((currentTime - dwellStartTime) >= dwellDuration && (lastUpdateTime - dwellStartTime) < dwellDuration) {
                    ArrayList<TouchPoint> inputInMms = new ArrayList<TouchPoint>(); // the recognizer stores templates in mm
                    ArrayList<TouchPoint> inputIn01 = new ArrayList<TouchPoint>();
                    for (int i = 0; i < FINGERS_LOCATION.length; i++) {
                        if(FINGERS_LOCATION[i] != null) {
                            inputInMms.add(new TouchPoint(
                                    FINGERS_LOCATION[i].x * viewWidthMm,
                                    FINGERS_LOCATION[i].y * viewHeightMm));
                            inputIn01.add(new TouchPoint(FINGERS_LOCATION[i].x, FINGERS_LOCATION[i].y));
                        }
                    }
                    if(debug) {
                        Log.v("TOUCHTOKENS", "input: "+inputInMms);
                        Log.v("TOUCHTOKENS", "MainActivity.RECOGNIZER: "+MainActivity.RECOGNIZER);
                    }
                    TokenTemplate tt = MainActivity.RECOGNIZER.recognize(inputInMms);
                    if(debug) {
                        Log.v("TOUCHTOKENS", "recognized as: "+tt);
                    }
                    if(tt != null) {
                        tokenDown = tt.getTokenID();
                        onTokenDown(tokenDown, inputIn01, viewWidthMm, viewHeightMm);
                    }
                }
                lastUpdateTime = System.currentTimeMillis();
            }
        }, 0, 25);

        tracePaint.setColor(Color.LTGRAY);
    }

    public abstract void onTokenDown(String tokenID, ArrayList<TouchPoint> points, double viewWidthMm, double viewHeightMm);

    public abstract void onTokenMoved(String tokenID, ArrayList<TouchPoint> points, double viewWidthMm, double viewHeightMm);

    public abstract void onTokenUp(String tokenID, ArrayList<TouchPoint> points, double viewWidthMm, double viewHeightMm);

    public void fingerDown(MotionEvent event, int index, int id) {
        FINGERS_LOCATION[id] = new TouchPoint(
                event.getX(index)/getWidth(),
                event.getY(index)/getHeight());
        for (int i = 0; i < FINGERS_LOCATION.length; i++) {
            FINGERS_LOCATION_AT_TIMER_START_TIME[i] = FINGERS_LOCATION[i];
        }
        dwellStartTime = System.currentTimeMillis();
    }

    public void fingerMove(MotionEvent event, int index, int id) {
        FINGERS_LOCATION[id] = new TouchPoint(
                event.getX(index)/getWidth(),
                event.getY(index)/getHeight());

        // sometimes, MT events are not delivered in the right order, avoid null exception:
        if(FINGERS_LOCATION_AT_TIMER_START_TIME[id] == null) {
            FINGERS_LOCATION_AT_TIMER_START_TIME[id] = FINGERS_LOCATION[id];
        }

        TouchPoint pointNow = new TouchPoint(
                FINGERS_LOCATION[id].x * viewWidthMm,
                FINGERS_LOCATION[id].y * viewHeightMm);
        TouchPoint pointAtTimerStartTime = new TouchPoint(
                FINGERS_LOCATION_AT_TIMER_START_TIME[id].x * viewWidthMm,
                FINGERS_LOCATION_AT_TIMER_START_TIME[id].y * viewHeightMm);

        double distance = pointNow.distance(pointAtTimerStartTime);

        boolean dwelling = distance < dwellTolerance;
        if(!dwelling) {
            for (int i = 0; i < FINGERS_LOCATION.length; i++) {
                FINGERS_LOCATION_AT_TIMER_START_TIME[i] = FINGERS_LOCATION[i];
            }
            dwellStartTime = System.currentTimeMillis();
        }

        ArrayList<TouchPoint> inputIn01 = new ArrayList<TouchPoint>();
        for (int i = 0; i < FINGERS_LOCATION.length; i++) {
            if(FINGERS_LOCATION[i] != null) {
                inputIn01.add(new TouchPoint(FINGERS_LOCATION[i].x, FINGERS_LOCATION[i].y));
            }
        }
        if(tokenDown != null) {
            onTokenMoved(tokenDown, inputIn01, viewWidthMm, viewHeightMm);
        }
    }

    public void fingerUp(MotionEvent event, int index, int id) {
        int cursorID = id;
        int fingerCount = 0;
        for (int i = 0; i < FINGERS_LOCATION.length; i++) {
            if(FINGERS_LOCATION[i] != null) {
                fingerCount++;
            }
        }
        if(fingerCount == 1) {
            dwellStartTime = -1;
            ArrayList<TouchPoint> inputIn01 = new ArrayList<TouchPoint>();
            for (int i = 0; i < FINGERS_LOCATION.length; i++) {
                if(FINGERS_LOCATION[i] != null) {
                    inputIn01.add(new TouchPoint(FINGERS_LOCATION[i].x, FINGERS_LOCATION[i].y));
                }
            }
            onTokenUp(tokenDown, inputIn01, viewWidthMm, viewHeightMm);
            tokenDown = null;
        } else {
            dwellStartTime = System.currentTimeMillis();
        }
        FINGERS_LOCATION[cursorID] = null;
        for (int i = 0; i < FINGERS_LOCATION.length; i++) {
            FINGERS_LOCATION_AT_TIMER_START_TIME[i] = FINGERS_LOCATION[i];
        }
    }

    public boolean onTouchEvent(MotionEvent event) {
        int index = MotionEventCompat.getActionIndex(event);
        int action = MotionEventCompat.getActionMasked(event);
        int id = MotionEventCompat.getPointerId(event, index);
        if(debug) {
            Log.v("TOUCHTOKENS", action+" ("+event.getX(index)+", "+event.getY(index)+") "+id+" - "+index);
        }
        switch (action) {
            case MotionEvent.ACTION_DOWN:
                fingerDown(event, index, id);
                break;
            case MotionEvent.ACTION_POINTER_DOWN:
                fingerDown(event, index, id);
                break;
            case MotionEvent.ACTION_MOVE:
                //fingerMove(event, index, id);
                int pointerCount = event.getPointerCount();
                for(int i = 0; i < pointerCount; ++i) {
                    index = i;
                    id = MotionEventCompat.getPointerId(event, index);
                    if(id < 10){
                        fingerMove(event, index, id);
                    }
                }
                break;
            case MotionEvent.ACTION_UP:
                fingerUp(event, index, id);
                break;
            case MotionEvent.ACTION_POINTER_UP:
                fingerUp(event, index, id);
                break;
            default:
                return true;
        }
        if(debug) {
            String points = "";
            for (int i = 0; i < FINGERS_LOCATION.length; i++) {
                if(FINGERS_LOCATION[i] != null) {
                    points += " "+FINGERS_LOCATION[i];
                }
            }
            Log.v("TOUCHTOKENS", "points="+points);
        }
        invalidate();
        return true;
    }

    protected void onDraw(Canvas canvas) {
        super.onDraw(canvas);
        viewWidthMm = getWidth() / MainActivity.SCREEN_RESOLUTION;
        viewHeightMm = getHeight() / MainActivity.SCREEN_RESOLUTION;

        for (int i = 0; i < FINGERS_LOCATION.length; i++) {
            if(FINGERS_LOCATION[i] != null) {
                canvas.drawCircle(
                        (int)(FINGERS_LOCATION[i].x * getWidth()),
                        (int)(FINGERS_LOCATION[i].y * getHeight()),
                        40, tracePaint);
            }
        }
    }


}
